/* colors */

html[data-theme="light"] {
    --pst-color-primary: rgb(19, 156, 90);
    --pst-color-secondary: rgb(231, 4, 136);
}

html[data-theme="dark"] {
    --pst-color-primary: rgb(19, 156, 90);
    --pst-color-secondary: rgb(231, 4, 136);
}


/* buttons */

.button>p>a {
    box-shadow: 0px 4px 14px -7px #999999;
    /* background-color: white; */
    border: 1px solid #bbbbbb;
    display: inline-block;
    cursor: pointer;
    color: #139C5A;
    padding: 1em 1em;
    text-align: center;
    text-decoration: none;
    font-size: 140%;
    margin: 2%;
    width: 40%;
}

.button>p>a:hover {
    border-color: rgb(19, 156, 90);
    color: rgb(231, 4, 136);
}

.button>p>a:active {
    position: relative;
    top: 1px;
}